##################################################################
# July 27 2013
# Jeff Goldsmith
#
# wrapper for doing variational bayes approach to penalized
# function on scalar regression with subject-level random effects
# and correlated errors
##################################################################

library(splines)
library(spam)

varbayes.pfsr = function(Y, fixef.form, id, data, Kt, alpha = .1){

  ## fixed and random effect design matrices
  W.des = model.matrix( fixef.form, data = data)
  Z.des = model.matrix( ~ 0 + as.factor(id) + (-1):as.factor(id))
  W.des = as.spam(W.des)
  Z.des = as.spam(Z.des)

  I = dim(Z.des)[2]
  D = dim(Y)[2]
  Ji = as.numeric(apply(Z.des, 2, sum))
  IJ = sum(Ji)
  p = dim(W.des)[2]

  ## bspline basis and penalty matrix
  Theta = bs(1:D, df=Kt, intercept=TRUE, degree=3)

  diff0 = diag(1, D, D)
  diff2 = matrix(rep(c(1,-2,1, rep(0, D-2)), D-2)[1:((D-2)*D)], D-2, D, byrow = TRUE)
  P0 = t(Theta) %*% t(diff0) %*% diff0 %*% Theta
  P2 = t(Theta) %*% t(diff2) %*% diff2 %*% Theta
  P.mat = alpha * P0 + (1-alpha) * P2

  SUBJ = factor(apply(Z.des %*% 1:dim(Z.des)[2], 1, sum))

  ## find first observation
  firstobs = rep(NA, length(unique(SUBJ)))
  for(i in 1:length(unique(SUBJ))){
    firstobs[i] = which(SUBJ %in% unique(SUBJ)[i])[1]
  }
  Wi = W.des[firstobs,]

  ## data organization; these computations only need to be done once
  Y.vec = as.vector(t(Y))
  IIP = kronecker(diag(1, I, I), P.mat)
  WIk = kronecker(Wi, diag(1, Kt, Kt))
  tWIW = t(WIk) %*% IIP %*% WIk
  tWI = t(WIk) %*% IIP


  ## initial estimation and hyperparameter choice
  mu.q.BZ = matrix(NA, nrow = Kt, ncol = I)
  for(subj in 1:length(unique(SUBJ))){
    mu.q.BZ[,subj] = solve(kronecker(Ji[subj], t(Theta) %*% Theta)) %*% t(kronecker(rep(1, Ji[subj]), Theta)) %*% 
                     (as.vector(t(Y[which(SUBJ == unique(SUBJ)[subj]),])))
    
  }
  vec.BZ = as.vector(mu.q.BZ)
  
  vec.BW = solve(tWIW) %*% tWI %*% vec.BZ
  mu.q.BW = matrix(vec.BW, Kt, p)

  Yhat = as.matrix(Z.des %*% t(mu.q.BZ) %*% t(Theta))
  varhat = var(as.vector(Y - Yhat))

  Psi = diag(varhat*IJ, D, D)
  v = IJ
  inv.sig = solve(Psi/v)

  Az = I*Kt / 2
  Bz = b.q.lambda.BZ = sum(diag((t(mu.q.BZ) - Wi %*% t(mu.q.BW)) %*% P.mat %*% t(t(mu.q.BZ) - Wi %*% t(mu.q.BW))))
  
  Aw = Kt / 2
  Bw = b.q.lambda.BW = sapply(1:p, function(u) max(1, sum(diag( t(mu.q.BW[,u]) %*% P.mat %*% (mu.q.BW[,u])))))


  sigma.q.BZ = vector("list", I)
  for(k in 1:p){
    sigma.q.BZ[[k]] = diag(1, Kt)
  }

  lpxq=c(0,1)
  j=2

  while(j<4 | (lpxq[j]-lpxq[j-1])>1.0E-1){
  
    ###############################################################
    ## update b-spline parameters for subject random effects
    ###############################################################

    for(subj in 1:length(unique(SUBJ))){
      
      t.designmat.Z = t(kronecker(rep(1, Ji[subj]), Theta))

      sigma.q.BZ[[subj]] = solve(t.designmat.Z %*% kronecker(diag(1, Ji[subj], Ji[subj]), inv.sig) %*% t(t.designmat.Z) + 
                         ((Az+I*Kt/2)/b.q.lambda.BZ) * P.mat)
      mu.q.BZ[,subj] = sigma.q.BZ[[subj]] %*% (t.designmat.Z %*% kronecker(diag(1, Ji[subj], Ji[subj]), inv.sig) %*% 
                                               (as.vector(t(Y[which(SUBJ == unique(SUBJ)[subj]),]))) + 
                          (((Az+I*Kt/2)/b.q.lambda.BZ) * P.mat) %*% mu.q.BW %*% t(Wi[subj,]) )
    
    }

    ranef.cur = as.matrix(Z.des %*% t(mu.q.BZ) %*% t(Theta))

    ###############################################################
    ## update b-spline parameters for fixed effects
    ###############################################################
    
    sigma.q.BW = solve(  (((Az+I*Kt/2)/b.q.lambda.BZ) * tWIW) + kronecker(diag((Aw+Kt/2)/b.q.lambda.BW), P.mat ))
    mu.q.BW = matrix( sigma.q.BW %*% (((Az+I*Kt/2)/b.q.lambda.BZ) * tWI %*% as.vector(mu.q.BZ)), nrow = Kt, ncol = p)

    beta.cur = t(mu.q.BW) %*% t(Theta)
 
    ###############################################################
    ## update inverse covariance matrix
    ###############################################################
    
    T.BZ.Z = Theta %*% mu.q.BZ %*% t(Z.des)

    mu.q.v = v + IJ
    mu.q.Psi = Psi + t(Y) %*% Y - 
               t(Y) %*% t(T.BZ.Z) -
               T.BZ.Z %*% Y +
               T.BZ.Z %*% t(T.BZ.Z) + 
               matrix(apply(sapply(1:I, function(u) Ji[u] * Theta %*% sigma.q.BZ[[u]] %*% t(Theta)), 1, sum), D, D)

    inv.sig = solve(mu.q.Psi/mu.q.v)

    ###############################################################
    ## update variance components
    ###############################################################

    ## lambda for fixed effects
    for(term in 1:dim(W.des)[2]){
      b.q.lambda.BW[term] = Bw[term] + .5 * (t(mu.q.BW[,term]) %*% P.mat %*% mu.q.BW[,term] + 
                                        sum(diag(P.mat %*% sigma.q.BW[(Kt*(term-1)+1):(Kt*term),(Kt*(term-1)+1):(Kt*term)])))
    }

    ## lambda for random effects
    trace.temp = sapply(1:dim(W.des)[2], function(p) sum(diag(P.mat %*% sigma.q.BW[(Kt*(p-1)+1):(Kt*p),(Kt*(p-1)+1):(Kt*p)])) )
    b.q.lambda.BZ = as.numeric(Bz + .5 * sum(sapply(1:I, function(u) {
    	                                  mu.q.BZ[,u] %*% P.mat %*% mu.q.BZ[,u] + sum(diag(P.mat %*% sigma.q.BZ[[u]])) -
    	                                  2 * Wi[u,] %*% t(mu.q.BW) %*% P.mat %*% mu.q.BZ[,u] +
    	                                  Wi[u,] %*% t(mu.q.BW) %*% P.mat %*% mu.q.BW %*% t(Wi[u,]) + 
    	                                  Wi[u,] %*% diag(trace.temp) %*% t(Wi[u,])
                                        })))
    
    curlpxq = .5 * sum(sapply(1:I, function(u) det(sigma.q.BZ[[u]]))) + 
              .5 * sum(sapply(1:p, function(u) det(sigma.q.BW[(Kt*(p-1)+1):(Kt*p),(Kt*(p-1)+1):(Kt*p)]))) -
              (Az + (I*Kt)/2) * log(b.q.lambda.BZ) - 
              (Aw + Kt/2) * sum(b.q.lambda.BW) -
              (v + IJ)/2 * log(det(mu.q.Psi))
    lpxq = c(lpxq, curlpxq)
    j=j+1

  }

  lpxq=lpxq[-(1:2)]

  ## compute CI for fixed effects
  beta.sd = beta.LB = beta.UB = matrix(NA, nrow = p, ncol = D)
  for(i in 1:p){
    beta.sd[i,] = sqrt(diag((Theta) %*% sigma.q.BW[(Kt*(i-1)+1):(Kt*i),(Kt*(i-1)+1):(Kt*i)] %*% t(Theta)))
    beta.LB[i,] = beta.cur[i,]-1.96*beta.sd[i,]
    beta.UB[i,] = beta.cur[i,]+1.96*beta.sd[i,]
  }

  ## convert objects from spam to matrix
  beta.cur = as.matrix(beta.cur)

  ## export fitted values
  Yhat.fixed = W.des %*% beta.cur
  Yhat = ranef.cur
  
  ## export various r2 values
  r2.f = 1 - (sum((Y - Yhat.fixed)^2)/(IJ*D)) / (sum((Y)^2)/(IJ*D))
  r2.fr = 1 - (sum((Y - Yhat)^2)/(IJ*D)) / (sum((Y)^2)/(IJ*D))
    
  ret = list(beta.cur, beta.UB, beta.LB, Yhat, r2.f, r2.fr)
  names(ret) = c("beta.pm", "beta.UB", "beta.LB", "Yhat", "r2.f", "r2.fr")
  
  ret

}
  

###############################################################
###############################################################
###############################################################
###############################################################
###############################################################
###############################################################